/*
 This file is part of the digitalSTROM ds485-core library.

 Copyright (C) 2010 digitalSTROM.org, Schlieren, Switzerland

 digitalSTROM ds485-core library is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation version 2
 of the License.

 digitalSTROM ds485-core library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with digitalSTROM ds485-core library; if not, write to the Free Software
 Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#include <string.h>
#include <stdlib.h>
#include <assert.h>

#include "digitalSTROM/ds485.h"
#include "digitalSTROM/ds485-const.v1.h"
#include "ds485-qp.v1.h"
#include "ds485-signals.h"
#include "unix/bsp.h"
#include "unix/callqueue.h"
#include "digitalSTROM/ds485-core.h"
#include "mmi.h"
#include "crc16_ccitt.h"
#include "bitset8.h"

Q_DEFINE_THIS_MODULE(ds485)

enum DS485Timeouts {
    /* transmit message timeout ticks */
    TX_MSG_TIMEOUT_TICKS                                = (100/BSP_MS_PER_TICKS),
    /* slave startup timeout ticks */
    SLAVE_STARTUP_TIMEOUT_TICKS                         = (30000/BSP_MS_PER_TICKS),
    /* solicit successor request timeout ticks */
    SOLICIT_SUCCESSOR_REQUEST_TIMEOUT_TICKS             = (200/BSP_MS_PER_TICKS),
    /* set address response timeout ticks */
    SET_ADDRESS_RESPONSE_TIMEOUT_TICKS                  = (100/BSP_MS_PER_TICKS),
    /* token min hold timeout ticks */
    TOKEN_MIN_HOLD_TIMEOUT_TICKS                        = (10/BSP_MS_PER_TICKS),
    /* token max hold timeout ticks */
    TOKEN_MAX_HOLD_TIMEOUT_TICKS                        = (75/BSP_MS_PER_TICKS),
    /* token long hold timeout ticks */
    TOKEN_LONG_HOLD_TIMEOUT_TICKS                       = (500/BSP_MS_PER_TICKS),
    /* find station timeout ticks */
    FIND_STATION_TIMEOUT_TICKS                          = (((2000/BSP_MS_PER_TICKS)+SOLICIT_SUCCESSOR_REQUEST_TIMEOUT_TICKS)),
    /* token hand over timeout ticks */
    TOKEN_HAND_OVER_TIMEOUT_TICKS                       = (300/BSP_MS_PER_TICKS),
    /* token missing timeout ticks in master mode */
    TOKEN_MISSING_MASTER_TIMEOUT_TICKS                  = (10000/BSP_MS_PER_TICKS),
    /* token missing timeout ticks in slave mode */
    TOKEN_MISSING_SLAVE_TIMEOUT_TICKS                   = (15000/BSP_MS_PER_TICKS),

    DSMAPI_RESPONSE_TIMEOUT_TICKS                       = (200/BSP_MS_PER_TICKS),
    GET_ADDRESS_REQUEST_TIMEOUT_TICKS                   = (200/BSP_MS_PER_TICKS),
    WAIT_FOR_SOLICIT_SUCCESSOR_RESPONSE_TIMEOUT_TICKS   = (250/BSP_MS_PER_TICKS),
    WAIT_FOR_SET_ADDRESS_REQUEST_TIMEOUT_TICKS          = (1000/BSP_MS_PER_TICKS)
};

/**
 * Returns true if destination address is broadcast
 */
#define DS485_IsRxMessage(sig_)                 (((sig_) >= DS485_RX_TOKEN_SIG) && ((sig_) <= DS485_RX_EVENT_SIG))

/* Active object class -----------------------------------------------------*/

typedef struct DS485Tag
{
    QActive super;
    QTimeEvt startupTimeEvt;
    QTimeEvt txMsgTimeEvt;
    QTimeEvt responseTimeEvt;
    QTimeEvt tokenMinTimeEvt;
    QTimeEvt tokenMaxTimeEvt;
    QTimeEvt findStationTimeEvt;
    QTimeEvt handOverTokenTimeEvt;
    QTimeEvt tokenMissingTimeEvt;
    QTimeEvt registerStationTimeEvt;
    QTimeEvt transmitDataTimeEvt;
    QEQueue deferredTxMessageQueue;
    dsid_t dsmid;
    uint8_t master;
    uint8_t deviceAddress;
    uint8_t successorDeviceAddress;
    uint8_t predecessorDeviceAddress;
    dsid_t busDsid[64];
    uint8_t usedAddresses[8];
    uint8_t pingDeviceAddress;
    uint8_t startupTimeEvtDisarmed;
    uint8_t txMsgTimeEvtDisarmed;
    uint8_t responseTimeEvtDisarmed;
    uint8_t random;
    uint8_t try;
    uint8_t findStation;
    uint8_t tokenMinHoldTimeElapsed;
    uint8_t tokenMaxHoldTimeElapsed;
    uint8_t tokenHandoverCount;
    uint8_t terminate;
    uint8_t messageSentAfterReceivingToken;
    QStateHandler superRegisterStationHistory;
    QStateHandler exitRegisterStationHistory;
    QStateHandler tryRegisterStationHistory;
    DS485Statistics statistics;
} DS485;

static dsid_t NO_DEVICE    = { 0 };
static int cmpAddress( DS485 *me, uint8_t addr, dsid_t* dsid )
{
    assert(addr < 64);
    return  memcmp(&me->busDsid[addr], dsid, sizeof(dsid_t));
}

static bool isAddressValid( DS485 *me, uint8_t addr )
{
    return  cmpAddress(me,addr,&NO_DEVICE) != 0;
}

static void registerAddress( DS485 *me, uint8_t addr, dsid_t* dsid )
{
    assert(addr < 64);
    if (!IsNullId(*dsid))
    {
        me->busDsid[addr]   = *dsid;

        MMI_DS485DsidRegisteredEvt  *evt    = Q_NEW( MMI_DS485DsidRegisteredEvt, MMI_DS485_DSID_REGISTRED_SIG );
        evt->dsid   = *dsid;
        QActive_postFIFO(AO_MMI, (QEvent*)evt);
    }
}

static void unregisterAddress( DS485 *me, uint8_t addr )
{
    assert(addr < 64);
    if (isAddressValid(me, addr))
    {
        MMI_DS485DsidRegisteredEvt  *evt    = Q_NEW( MMI_DS485DsidRegisteredEvt, MMI_DS485_DSID_DEREGISTRED_SIG );
        evt->dsid   = me->busDsid[addr];
        QActive_postFIFO(AO_MMI, (QEvent*)evt);
        me->busDsid[addr]   = NO_DEVICE;
    }
}

static void clearAdresses( DS485 *me )
{
    memset(me->busDsid, 0, sizeof(me->busDsid));
}

static void checkForMissingBusMembers( DS485 *me )
{
    int i;
    for( i = 0; i < sizeof(me->busDsid)/sizeof(me->busDsid[0]); i++ ){
        if( IsBitSet8Cleared(me->usedAddresses,i) && isAddressValid( me, i ) ){
            unregisterAddress( me, i );
        }
    }
}

static QState DS485_initial (DS485 *me, QEvent const *e);
static QState DS485_wait_for_configuration(DS485 *me, QEvent const *e);
static QState DS485_top (DS485 *me, QEvent const *e);
static QState DS485_startup(DS485 *me, QEvent const *e);
static QState DS485_send_dsid(DS485 *me, QEvent const *e);

static QState DS485_startup_slave(DS485 *me, QEvent const *e);
static QState DS485_wait_for_solicit_successor_request(DS485 *me, QEvent const *e);
static QState DS485_wait_for_solicit_successor_request_short(DS485 *me, QEvent const *e);
static QState DS485_wait_for_solicit_successor_request_long(DS485 *me, QEvent const *e);
static QState DS485_send_solicit_response(DS485 *me, QEvent const *e);
static QState DS485_wait_for_set_address_request(DS485 *me, QEvent const *e);
static QState DS485_wait_for_set_successor_address_request(DS485 *me, QEvent const *e);

static QState DS485_startup_master(DS485 *me, QEvent const *e);
static QState DS485_send_solicit_successor_request_short(DS485 *me, QEvent const *e);

static QState DS485_register_station(DS485 *me, QEvent const *e);
static QState DS485_ping_device_address(DS485 *me, QEvent const *e);
static QState DS485_send_set_address_request(DS485 *me, QEvent const *e);
static QState DS485_send_set_successor_address_request(DS485 *me, QEvent const *e);
static QState DS485_send_set_successor_address_request_to_predecessor(DS485 *me, QEvent const *e);

static QState DS485_operational(DS485 *me, QEvent const *e);

static QState DS485_holding_token(DS485 *me, QEvent const *e);
static QState DS485_find_station(DS485 *me, QEvent const *e);
static QState DS485_wait_for_token_min_hold_time(DS485 *me, QEvent const *e);
static QState DS485_transmit_data(DS485* me, QEvent const *e);
static QState DS485_wait_for_acknowledge(DS485* me, QEvent const *e);

static QState DS485_no_token(DS485 *me, QEvent const *e);
static QState DS485_hand_over_token(DS485 *me, QEvent const *e);
static QState DS485_send_token(DS485 *me, QEvent const *e);
static QState DS485_find_successor_device(DS485 *me, QEvent const *e);

static QState DS485_stopped(DS485 *me, QEvent const *e);

static void DS485_SendToken(DS485 *me, uint8_t dest, uint8_t src);
static void DS485_SendMessage(DS485 *me, DS485MessageEvt *msgEvt);
static void DS485_SendMessageEx(DS485 *me, DS485MessageEvt *msgEvt);
static void DS485_SendSetAddressResponse(DS485 *me, uint8_t dest);
static void DS485_SendSetSuccessorAddressResponse(DS485 *me, uint8_t dest);
static void DS485_PublishDeviceAddress(uint8_t deviceAddress);
static void DS485_ChangeToMaster(DS485 *me);
static bool DS485_GetNextSuccessorAddress(DS485 *me);
static void DS485_ResetStatistics(DS485 *me);
static void DS485_RequestStatistics(DS485 *me, DS485ReqStatisticsEvt const * reqEvt);

static uint8_t DS485_getBusIdByDsid( DS485 *me, dsid_t *dsid );
static dsid_t DS485_getDsidByBusId( DS485 *me, uint8_t busId );

/* Local objects -----------------------------------------------------------*/
static DS485 l_ds485;   /* the single instance of the dS485 active object */
static QEvent const * l_deferredTxMessageQueueSto[8];   /* event queue storage */
static QEvent const * l_ds485TxMessageQueueSto[8];    /* event queue storage */
static const QEvent l_ds485DeviceLostEvt = { MMI_DS485_DEVICE_LOST_SIG, 0 };
static QEQueue l_ds485TxMessageQueue;   /* transmit message queue */
static const MMI_DS485StateEvt l_joinEvt = { { MMI_DS485_STATE_SIG, 0 }, MMI_DS485_JOINING };

/* Global-scope objects ----------------------------------------------------*/
QActive * const AO_DS485 = (QActive *) &l_ds485; /* "opaque" AO pointer */

/*..........................................................................*/
void DS485_ctor(void)
{
    DS485 *me = &l_ds485;
    QActive_ctor(&me->super, (QStateHandler)&DS485_initial);
    /* initialize message event transmit queue */
    QEQueue_init(&l_ds485TxMessageQueue, l_ds485TxMessageQueueSto, Q_DIM(l_ds485TxMessageQueueSto));
    QEQueue_init(&me->deferredTxMessageQueue, l_deferredTxMessageQueueSto, Q_DIM(l_deferredTxMessageQueueSto));
    /* initialize time events */
    QTimeEvt_ctor(&me->startupTimeEvt, STARTUP_TIMEOUT_SIG);
    QTimeEvt_ctor(&me->txMsgTimeEvt, TX_MSG_TIMEOUT_SIG);
    QTimeEvt_ctor(&me->responseTimeEvt, RESPONSE_TIMEOUT_SIG);
    QTimeEvt_ctor(&me->tokenMinTimeEvt, TOKEN_MIN_HOLD_TIMEOUT_SIG);
    QTimeEvt_ctor(&me->tokenMaxTimeEvt, TOKEN_MAX_HOLD_TIMEOUT_SIG);
    QTimeEvt_ctor(&me->findStationTimeEvt, FIND_STATION_TIMEOUT_SIG);
    QTimeEvt_ctor(&me->handOverTokenTimeEvt, HAND_OVER_TOKEN_TIMEOUT_SIG);
    QTimeEvt_ctor(&me->tokenMissingTimeEvt, TOKEN_MISSING_TIMEOUT_SIG);
    QTimeEvt_ctor(&me->registerStationTimeEvt, REGISTER_STATION_TIMEOUT_SIG);
    QTimeEvt_ctor(&me->transmitDataTimeEvt, TRANSMIT_DATA_TIMEOUT_SIG);
    me->terminate = FALSE;
}

/**
 * Create DS485 message
 */
DS485MessageEvt * DS485_CreateMessage(uint8_t dest, uint8_t src, uint8_t msgCount, uint8_t cmd, uint8_t payloadLength)
{
    uint8_t* message;
    /* allocate new message event */
    DS485MessageEvt * msgEvt = Q_NEW(DS485MessageEvt, DS485_TX_MSG_SIG);
    /* build header */
    message = msgEvt->message;
    message[0] = DS485_START_BYTE;
    if (dest == DS485_BROADCAST) {
        message[1] = DS485_BROADCAST_MASK | DS485_TYPEFLAG_MASK;
    }
    else {
        message[1] = (dest << 2) | DS485_TYPEFLAG_MASK;
    }
    message[2] = (src << 2) | (msgCount & 0x3);
    message[3] = (cmd << 4) | (payloadLength & 0xF);
    /* set message size = message length + 4 header + 2 CRC-16 */
    msgEvt->size = payloadLength + 6;
    return msgEvt;
}

/**
 * Create DS485 data message
 */
DS485MessageEvt * DS485_CreateDataMessage(const dsid_t *dest, DS485CommandType type, uint8_t payloadLength)
{
    uint8_t* message;
    uint8_t cmd;
    Q_ASSERT(payloadLength <= DS485_SHORT_PAYLOAD_SIZE);
    /* allocate new message event */
    DS485MessageEvt * msgEvt = Q_NEW(DS485MessageEvt, DS485_TX_MSG_SIG);
    /* build header */
    message = msgEvt->message;
    message[0] = DS485_START_BYTE;
    message[1] = DS485_TYPEFLAG_MASK;           /* destination */
    message[2] = 0;                             /* source */

    switch (type) {
    default:
    case DS485_TYPE_REQ:
        message[3] = (DS485_REQUEST << 4) | (payloadLength & 0xF);
        break;
    case DS485_TYPE_RES:
        message[3] = (DS485_RESPONSE << 4) | (payloadLength & 0xF);
        break;
    case DS485_TYPE_EVT:
        message[3] = (DS485_EVENT << 4) | (payloadLength & 0xF);
        break;
    }

    /* set message size = message length + 4 header + 2 CRC-16 */
    msgEvt->size = payloadLength + DS485_HEADER_CRC_SIZE;
    msgEvt->destinationAddr = *dest;
    return msgEvt;
}

DS485MessageEvt const * DS485_GetNextTxMessage()
{
    return (DS485MessageEvt const *) QEQueue_get(&l_ds485TxMessageQueue);
}

/*..........................................................................*/
static QState DS485_initial(DS485 *me, QEvent const *e)
{
    (void) e; /* suppress the compiler warning about unused parameter */
    /* QS dictionary entries */
    QS_OBJ_DICTIONARY(&l_ds485);

    QS_FUN_DICTIONARY(&DS485_initial);
    QS_FUN_DICTIONARY(&DS485_wait_for_configuration);
    QS_FUN_DICTIONARY(&DS485_top);
    QS_FUN_DICTIONARY(&DS485_startup);
    QS_FUN_DICTIONARY(&DS485_send_dsid);

    QS_FUN_DICTIONARY(&DS485_startup_slave);
    QS_FUN_DICTIONARY(&DS485_wait_for_solicit_successor_request);
    QS_FUN_DICTIONARY(&DS485_wait_for_solicit_successor_request_short);
    QS_FUN_DICTIONARY(&DS485_wait_for_solicit_successor_request_long);
    QS_FUN_DICTIONARY(&DS485_send_solicit_response);
    QS_FUN_DICTIONARY(&DS485_wait_for_set_address_request);
    QS_FUN_DICTIONARY(&DS485_wait_for_set_successor_address_request);

    QS_FUN_DICTIONARY(&DS485_startup_master);
    QS_FUN_DICTIONARY(&DS485_send_solicit_successor_request_short);

    QS_FUN_DICTIONARY(&DS485_register_station);
    QS_FUN_DICTIONARY(&DS485_ping_device_address);
    QS_FUN_DICTIONARY(&DS485_send_set_address_request);
    QS_FUN_DICTIONARY(&DS485_send_set_successor_address_request);
    QS_FUN_DICTIONARY(&DS485_send_set_successor_address_request_to_predecessor);

    QS_FUN_DICTIONARY(&DS485_operational);

    QS_FUN_DICTIONARY(&DS485_holding_token);
    QS_FUN_DICTIONARY(&DS485_find_station);
    QS_FUN_DICTIONARY(&DS485_wait_for_token_min_hold_time);
    QS_FUN_DICTIONARY(&DS485_transmit_data);
    QS_FUN_DICTIONARY(&DS485_wait_for_acknowledge);

    QS_FUN_DICTIONARY(&DS485_no_token);
    QS_FUN_DICTIONARY(&DS485_hand_over_token);
    QS_FUN_DICTIONARY(&DS485_send_token);

    QS_FUN_DICTIONARY(&DS485_stopped);

    QS_SIG_DICTIONARY(STARTUP_TIMEOUT_SIG, 0);
    QS_SIG_DICTIONARY(TX_MSG_TIMEOUT_SIG, 0);
    QS_SIG_DICTIONARY(RESPONSE_TIMEOUT_SIG, 0);
    QS_SIG_DICTIONARY(TOKEN_MIN_HOLD_TIMEOUT_SIG, 0);
    QS_SIG_DICTIONARY(TOKEN_MAX_HOLD_TIMEOUT_SIG, 0);
    QS_SIG_DICTIONARY(FIND_STATION_TIMEOUT_SIG, 0);
    QS_SIG_DICTIONARY(HAND_OVER_TOKEN_TIMEOUT_SIG, 0);
    QS_SIG_DICTIONARY(TOKEN_MISSING_TIMEOUT_SIG, 0);
    QS_SIG_DICTIONARY(REGISTER_STATION_TIMEOUT_SIG, 0);
    QS_SIG_DICTIONARY(TRANSMIT_DATA_TIMEOUT_SIG, 0);

    QS_SIG_DICTIONARY(DS485_RX_TOKEN_SIG, 0);
    QS_SIG_DICTIONARY(DS485_RX_SOLICIT_SUCCESSOR_REQUEST_LONG_SIG, 0);
    QS_SIG_DICTIONARY(DS485_RX_SOLICIT_SUCCESSOR_REQUEST_SHORT_SIG, 0);
    QS_SIG_DICTIONARY(DS485_RX_SOLICIT_SUCCESSOR_RESPONSE_SIG, 0);
    QS_SIG_DICTIONARY(DS485_RX_GET_ADDRESS_REQUEST_SIG, 0);
    QS_SIG_DICTIONARY(DS485_RX_GET_ADDRESS_RESPONSE_SIG, 0);
    QS_SIG_DICTIONARY(DS485_RX_SET_DEVICE_ADDRESS_REQUEST_SIG, 0);
    QS_SIG_DICTIONARY(DS485_RX_SET_DEVICE_ADDRESS_RESPONSE_SIG, 0);
    QS_SIG_DICTIONARY(DS485_RX_SET_SUCCESSOR_ADDRESS_REQUEST_SIG, 0);
    QS_SIG_DICTIONARY(DS485_RX_SET_SUCCESSOR_ADDRESS_RESPONSE_SIG, 0);
    QS_SIG_DICTIONARY(DS485_RX_REQUEST_SIG, 0);
    QS_SIG_DICTIONARY(DS485_RX_RESPONSE_SIG, 0);
    QS_SIG_DICTIONARY(DS485_RX_ACKNOWLEDGE_SIG, 0);
    QS_SIG_DICTIONARY(DS485_RX_BUSY_SIG, 0);
    QS_SIG_DICTIONARY(DS485_RX_DATA_SIG, 0);
    QS_SIG_DICTIONARY(DS485_RX_PING_SIG, 0);
    QS_SIG_DICTIONARY(DS485_RX_EVENT_SIG, 0);

    QS_SIG_DICTIONARY(DS485_TX_MSG_SIG, 0);
    QS_SIG_DICTIONARY(DS485_TX_RESPONSE_SIG, 0);

#if 0
    QS_FILTER_SM_OBJ(me);
    QS_FILTER_AO_OBJ(me);

    QS_FILTER_EQ_OBJ(&me->deferredTxMessageQueue);
    QS_FILTER_EQ_OBJ(&l_ds485TxMessageQueue);
#endif

    /* subscribe events */
    QActive_subscribe((QActive *)me, CONFIG_INIT_SIG);

    return Q_TRAN(&DS485_wait_for_configuration);
}

/*..........................................................................*/
static QState DS485_wait_for_configuration(DS485 *me, QEvent const *e)
{
    switch (e->sig) {
    case CONFIG_INIT_SIG: {
        DatamodelConfigEvt const * config = (DatamodelConfigEvt const *)e;
        /* save dSM ID */
        me->dsmid = config->dsmid;
        /* subscribe events */
        QActive_subscribe((QActive *) me, SYSTEM_TERM_SIG);
        return Q_TRAN(&DS485_startup);
    }
    }
    return Q_SUPER(&QHsm_top);
}

/*..........................................................................*/
static QState DS485_top(DS485 *me, QEvent const *e)
{
    switch (e->sig) {
    case DS485_TX_MSG_SIG:
        /* defer transmit message event */
        if (QEQueue_getNFree(&me->deferredTxMessageQueue) != 0) {
            QActive_defer((QActive *)me, &me->deferredTxMessageQueue, e);
        }
        return Q_HANDLED();
    case DS485_REQ_STATISTICS_SIG:
        DS485_RequestStatistics(me, (DS485ReqStatisticsEvt const *)e);
        return Q_HANDLED();
    case DS485_RESET_STATISTICS_SIG:
        DS485_ResetStatistics(me);
        return Q_HANDLED();
    case SYSTEM_TERM_SIG:
        return Q_TRAN(&DS485_stopped);
    case DS485_REQ_BUS_MEMBERS_SIG:
    {
        DS485_DsidResponseEvt    *evt    = Q_NEW(DS485_DsidResponseEvt, DS485_RES_BUS_MEMBERS_SIG);
        memset( evt->dsid, 0, sizeof(evt->dsid) );
        ccSendResponse( e, (QEvent*)evt );
        return Q_HANDLED();
    }
    }
    return Q_SUPER(&QHsm_top);
}
/*..........................................................................*/
static QState DS485_startup(DS485 *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG: {
        /* generate a random time between 2000 and 3000 ms in steps of 2 ms */
        uint16_t time = 2000 + ((rand() % 500 + 1) * 2);
        QTimeEvt_postIn(&me->startupTimeEvt, (QActive *)me, (time / BSP_MS_PER_TICKS));
        /* reset used addresses */
        memset(me->usedAddresses, 0, sizeof(me->usedAddresses));
        clearAdresses(me);
        /* reset predecessor device address */
        me->predecessorDeviceAddress = DS485_INVALID_ADDRESS;
        return Q_HANDLED();
    }
    case Q_EXIT_SIG:
        me->startupTimeEvtDisarmed = QTimeEvt_disarm(&me->startupTimeEvt);
        return Q_HANDLED();
#if 0
    case DS485_RX_TOKEN_SIG:
        /* fall through */
#endif
    case DS485_RX_SOLICIT_SUCCESSOR_REQUEST_LONG_SIG:
        /* fall through */
    case DS485_RX_SOLICIT_SUCCESSOR_REQUEST_SHORT_SIG:
        /* fall through */
    case DS485_RX_GET_ADDRESS_REQUEST_SIG:
        /* fall through */
    case DS485_RX_SET_DEVICE_ADDRESS_REQUEST_SIG:
        /* fall through */
    case DS485_RX_SET_SUCCESSOR_ADDRESS_REQUEST_SIG:
        /* fall through */
    case DS485_RX_RESPONSE_SIG:
        /* fall through */
    case DS485_RX_REQUEST_SIG:
        /* fall through */
    case DS485_RX_EVENT_SIG:
        /* any message received -> startup as slave */
        return Q_TRAN(&DS485_startup_slave);
    case STARTUP_TIMEOUT_SIG:
        return Q_TRAN(&DS485_send_dsid);
    }
    return Q_SUPER(&DS485_top);
}

/*..........................................................................*/
static QState DS485_send_dsid(DS485 *me, QEvent const *e)
{
    switch (e->sig)
    {
    case Q_ENTRY_SIG:
    {
        /* send solicit successor request short with own dSID */
        DS485MessageEvt* txMsgEvt = DS485_CreateMessage(0, 0, 0, DS485_SOLICIT_SUCCESSOR_REQUEST_SHORT, sizeof(me->dsmid));
        memcpy(&txMsgEvt->message[DS485_DATA_OFFSET], &me->dsmid, sizeof(me->dsmid));
        DS485_SendMessage(me, txMsgEvt);
        QTimeEvt_postIn(&me->txMsgTimeEvt, (QActive *)me, TX_MSG_TIMEOUT_TICKS);
        return Q_HANDLED();
    }
    case DS485_RX_SOLICIT_SUCCESSOR_REQUEST_SHORT_SIG:
    {
        me->txMsgTimeEvtDisarmed = QTimeEvt_disarm(&me->txMsgTimeEvt);
        /* compare dSID */
        if (IsEqualId(((DS485MessageEvt const *) e)->message[DS485_DATA_OFFSET], me->dsmid)) {
            /* solicit successor request short with own dSID -> start up as master */
            return Q_TRAN(&DS485_startup_master);
        }
        /* solicit successor request short with another dSID -> start up as slave */
        return Q_TRAN(&DS485_startup_slave);
    }
    case DS485_RX_TOKEN_SIG:
        /* fall through */
    case DS485_RX_SET_DEVICE_ADDRESS_REQUEST_SIG:
        /* fall through */
    case DS485_RX_SET_SUCCESSOR_ADDRESS_REQUEST_SIG:
        /* fall through */
    case DS485_RX_RESPONSE_SIG:
        /* fall through */
    case DS485_RX_REQUEST_SIG:
        /* fall through */
    case DS485_RX_EVENT_SIG:
        /* any message received -> startup as slave */
        return Q_TRAN(&DS485_startup_slave);
    case TX_MSG_TIMEOUT_SIG:
        /* timeout occurred -> restart from scratch */
        return Q_TRAN(&DS485_startup);
    }
    return Q_SUPER(&DS485_top);
}

/*..........................................................................*/
static QState DS485_startup_slave(DS485 *me, QEvent const *e)
{
    switch (e->sig)
    {
    case Q_ENTRY_SIG:
        me->master = FALSE;
        /* TODO: check if still necessary */
        me->deviceAddress = DS485_INTERMEDIATE_ADDRESS;
        QTimeEvt_postIn(&me->startupTimeEvt, (QActive *)me, SLAVE_STARTUP_TIMEOUT_TICKS);
        return Q_HANDLED();
    case Q_EXIT_SIG:
        QTimeEvt_disarm(&me->startupTimeEvt);
        return Q_HANDLED();
    case Q_INIT_SIG:
        return Q_TRAN(&DS485_wait_for_solicit_successor_request);
    case STARTUP_TIMEOUT_SIG:
        /* timeout occurred -> restart from scratch */
        return Q_TRAN(&DS485_startup);
    }
    return Q_SUPER(&DS485_top);
}

/*..........................................................................*/
static QState DS485_wait_for_solicit_successor_request(DS485 *me, QEvent const *e)
{
    switch (e->sig)
    {
    case DS485_RX_SOLICIT_SUCCESSOR_REQUEST_SHORT_SIG:
        return Q_TRAN(&DS485_wait_for_solicit_successor_request_short);
    case DS485_RX_SOLICIT_SUCCESSOR_REQUEST_LONG_SIG:
        return Q_TRAN(&DS485_wait_for_solicit_successor_request_long);
    }
    return Q_SUPER(&DS485_startup_slave);
}

/*..........................................................................*/
static QState DS485_wait_for_solicit_successor_request_short(DS485 *me, QEvent const *e)
{
    switch (e->sig)
    {
    case Q_ENTRY_SIG:
        /* generate a random number between 1 and 64 */
        me->random = (rand() % 64) + 1;
        QActive_postFIFO(AO_MMI, (QEvent *)&l_joinEvt);
        return Q_HANDLED();
    case DS485_RX_SOLICIT_SUCCESSOR_REQUEST_SHORT_SIG:
        if (--me->random == 0) {
            return Q_TRAN(&DS485_send_solicit_response);
        }
        return Q_HANDLED();
    case DS485_RX_SOLICIT_SUCCESSOR_REQUEST_LONG_SIG:
        return Q_TRAN(&DS485_wait_for_solicit_successor_request_long);
    }
    return Q_SUPER(&DS485_startup_slave);
}

/*..........................................................................*/
static QState DS485_wait_for_solicit_successor_request_long(DS485 *me, QEvent const *e)
{
    switch (e->sig)
    {
    case Q_ENTRY_SIG:
        /* generate a random number between 1 and 10 */
        me->random = (rand() % 10) + 1;
        QActive_postFIFO(AO_MMI, (QEvent *)&l_joinEvt);
        return Q_HANDLED();
    case DS485_RX_SOLICIT_SUCCESSOR_REQUEST_LONG_SIG:
        if (--me->random == 0) {
            return Q_TRAN(&DS485_send_solicit_response);
        }
        return Q_HANDLED();
    case DS485_RX_SOLICIT_SUCCESSOR_REQUEST_SHORT_SIG:
        return Q_TRAN(&DS485_wait_for_solicit_successor_request_short);
    }
    return Q_SUPER(&DS485_startup_slave);
}

/*..........................................................................*/
static QState DS485_send_solicit_response(DS485 *me, QEvent const *e)
{
    switch (e->sig)
    {
    case Q_ENTRY_SIG:
    {
        /* send solicit response */
        DS485MessageEvt* txMsgEvt = DS485_CreateMessage(0, me->deviceAddress, 0, DS485_SOLICIT_SUCCESSOR_RESPONSE, sizeof(me->dsmid));
        DS485_SetDSID(txMsgEvt, me->dsmid);
        DS485_SendMessage(me, txMsgEvt);
        QTimeEvt_postIn(&me->txMsgTimeEvt, (QActive *)me, WAIT_FOR_SOLICIT_SUCCESSOR_RESPONSE_TIMEOUT_TICKS);
        return Q_HANDLED();
    }
    case Q_EXIT_SIG:
        QTimeEvt_disarm(&me->txMsgTimeEvt);
        return Q_HANDLED();
    case DS485_RX_SOLICIT_SUCCESSOR_RESPONSE_SIG:
    {
        DS485MessageEvt const * rxMsgEvt = (DS485MessageEvt const *) e;
        if (IsEqualId(rxMsgEvt->message[DS485_DATA_OFFSET], me->dsmid)) {
            /* set intermediate address */
            me->deviceAddress = DS485_INTERMEDIATE_ADDRESS;
            return Q_TRAN(&DS485_wait_for_set_address_request);
        }
        return Q_TRAN(&DS485_wait_for_solicit_successor_request);
    }
    case TX_MSG_TIMEOUT_SIG:
        /* timeout occurred */
        return Q_TRAN(&DS485_wait_for_solicit_successor_request);
    }
    return Q_SUPER(&DS485_startup_slave);
}

/*..........................................................................*/
static QState DS485_wait_for_set_address_request(DS485 *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG:
        QTimeEvt_postIn(&me->txMsgTimeEvt, (QActive *)me, WAIT_FOR_SET_ADDRESS_REQUEST_TIMEOUT_TICKS);
        return Q_HANDLED();
    case Q_EXIT_SIG:
        QTimeEvt_disarm(&me->txMsgTimeEvt);
        return Q_HANDLED();
    case DS485_RX_SET_DEVICE_ADDRESS_REQUEST_SIG:
    {
        DS485MessageEvt const * rxMsgEvt = (DS485MessageEvt const *) e;
        /* save my device address */
        me->deviceAddress = rxMsgEvt->message[DS485_DATA_OFFSET];
        /* send set device address response */
        DS485MessageEvt * txMsgEvt = DS485_CreateMessage(DS485_GetSource(rxMsgEvt), me->deviceAddress, 0,
                DS485_SET_DEVICE_ADDRESS_RESPONSE, sizeof(me->dsmid));
        DS485_SetDSID(txMsgEvt, me->dsmid);
        DS485_SendMessage(me, txMsgEvt);
        /* wait for set successor address request */
        return Q_TRAN(&DS485_wait_for_set_successor_address_request);
    }
    case DS485_RX_SOLICIT_SUCCESSOR_REQUEST_SHORT_SIG:
        return Q_TRAN(&DS485_wait_for_solicit_successor_request_short);
    case DS485_RX_SOLICIT_SUCCESSOR_REQUEST_LONG_SIG:
        return Q_TRAN(&DS485_wait_for_solicit_successor_request_long);
    }
    return Q_SUPER(&DS485_startup_slave);
}

/*..........................................................................*/
static QState DS485_wait_for_set_successor_address_request(DS485 *me, QEvent const *e)
{
    switch (e->sig)
    {
    case DS485_RX_SET_SUCCESSOR_ADDRESS_REQUEST_SIG:
    {
        DS485MessageEvt const * rxMsgEvt = (DS485MessageEvt const *) e;
        if (DS485_GetDestination(rxMsgEvt) == me->deviceAddress) {
            /* save successor device address */
            me->successorDeviceAddress = rxMsgEvt->message[DS485_DATA_OFFSET];
            /* send set successor address response */
            DS485MessageEvt * txMsgEvt = DS485_CreateMessage(DS485_GetSource(rxMsgEvt), me->deviceAddress, 0,
                    DS485_SET_SUCCESSOR_ADDRESS_RESPONSE, sizeof(me->dsmid));
            DS485_SetDSID(txMsgEvt, me->dsmid);
            DS485_SendMessage(me, txMsgEvt);
            /* go to operational state */
            return Q_TRAN(&DS485_operational);
        }
        return Q_HANDLED();
    }
    case DS485_RX_SET_DEVICE_ADDRESS_REQUEST_SIG:
    {
        DS485MessageEvt const * rxMsgEvt = (DS485MessageEvt const *) e;
        /* save my device address */
        me->deviceAddress = rxMsgEvt->message[DS485_DATA_OFFSET];
        /* send set device address response */
        DS485MessageEvt * txMsgEvt = DS485_CreateMessage(DS485_GetSource(rxMsgEvt), me->deviceAddress, 0,
                DS485_SET_DEVICE_ADDRESS_RESPONSE, sizeof(me->dsmid));
        DS485_SetDSID(txMsgEvt, me->dsmid);
        DS485_SendMessage(me, txMsgEvt);
        /* wait for set successor address request */
        return Q_TRAN(&DS485_wait_for_set_successor_address_request);
    }
    case DS485_RX_SOLICIT_SUCCESSOR_REQUEST_SHORT_SIG:
        return Q_TRAN(&DS485_wait_for_solicit_successor_request_short);
    case DS485_RX_SOLICIT_SUCCESSOR_REQUEST_LONG_SIG:
        return Q_TRAN(&DS485_wait_for_solicit_successor_request_long);
    }
    return Q_SUPER(&DS485_startup_slave);
}

/*..........................................................................*/
static QState DS485_startup_master(DS485 *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG:
        me->master = TRUE;
        SetBitSet8(me->usedAddresses, 0);
        me->busDsid[0]  = me->dsmid;
        me->predecessorDeviceAddress = 0;
        me->random = 0;
        return Q_HANDLED();
    case Q_INIT_SIG:
        return Q_TRAN(&DS485_send_solicit_successor_request_short);
    case DS485_RX_SET_SUCCESSOR_ADDRESS_REQUEST_SIG:
    {
        DS485MessageEvt const * rxMsgEvt = (DS485MessageEvt const *) e;
        if (DS485_GetDestination(rxMsgEvt) == me->deviceAddress) {
            /* save successor device address */
            me->successorDeviceAddress = rxMsgEvt->message[DS485_DATA_OFFSET];
            /* send set successor address response */
            DS485MessageEvt * txMsgEvt = DS485_CreateMessage(DS485_GetSource(rxMsgEvt), me->deviceAddress, 0,
                    DS485_SET_SUCCESSOR_ADDRESS_RESPONSE, sizeof(me->dsmid));
            DS485_SetDSID(txMsgEvt, me->dsmid);
            DS485_SendMessage(me, txMsgEvt);
        }
        return Q_HANDLED();
    }
    case TX_MSG_TIMEOUT_SIG:
        me->txMsgTimeEvtDisarmed = TRUE;
        return Q_HANDLED();
    }
    return Q_SUPER(&QHsm_top);
}

/*..........................................................................*/
static QState DS485_send_solicit_successor_request_short(DS485 *me, QEvent const *e)
{
    switch (e->sig)
    {
    case Q_ENTRY_SIG:
    {
        DS485MessageEvt* txMsgEvt;
        /* increment counter */
        ++me->random;
        /* send solicit successor request short */
        txMsgEvt = DS485_CreateMessage(0, 0, 0, DS485_SOLICIT_SUCCESSOR_REQUEST_SHORT, sizeof(me->dsmid));
        DS485_SetDSID(txMsgEvt, me->dsmid);
        DS485_SendMessage(me, txMsgEvt);
        QTimeEvt_postIn(&me->startupTimeEvt, (QActive *)me, SOLICIT_SUCCESSOR_REQUEST_TIMEOUT_TICKS);
        return Q_HANDLED();
    }
    case Q_EXIT_SIG:
        QTimeEvt_disarm(&me->startupTimeEvt);
        return Q_HANDLED();
    case DS485_RX_SOLICIT_SUCCESSOR_REQUEST_SHORT_SIG:
        /* skip own request */
        return Q_HANDLED();
    case DS485_RX_SOLICIT_SUCCESSOR_RESPONSE_SIG:
        /* store history */
        me->superRegisterStationHistory = (QStateHandler) &DS485_startup_master;
        me->exitRegisterStationHistory = (QStateHandler) &DS485_send_solicit_successor_request_short;
        return Q_TRAN(&DS485_register_station);
    case STARTUP_TIMEOUT_SIG:
        if (me->random < 128) {
            return Q_TRAN(&DS485_send_solicit_successor_request_short);
        }
        return Q_TRAN(&DS485_operational);
    }
    return Q_SUPER(&DS485_startup_master);
}

/*..........................................................................*/
static QState DS485_register_station(DS485 *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG:
        me->try = 0;
        return Q_HANDLED();
    case Q_INIT_SIG:
        /* initialize device address to ping */
        me->pingDeviceAddress = DS485_MIN_SLAVE_ADDRESS;
        return Q_TRAN(&DS485_ping_device_address);
    case REGISTER_STATION_TIMEOUT_SIG:
        if (me->try < 2) {
            /* retry again -> go to try history */
            return Q_TRAN(me->tryRegisterStationHistory);
        }
        /* exit to history */
        return Q_TRAN(me->exitRegisterStationHistory);
    }
    return Q_SUPER(me->superRegisterStationHistory);
}

/*..........................................................................*/
static QState DS485_ping_device_address(DS485 *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG: {
        /* find next unused slave address */
        while (me->pingDeviceAddress <= DS485_MAX_SLAVE_ADDRESS) {
            if (IsBitSet8Cleared(me->usedAddresses, me->pingDeviceAddress)) {
                break;
            }
            ++me->pingDeviceAddress;
        }
        Q_ASSERT(me->pingDeviceAddress <= DS485_MAX_SLAVE_ADDRESS);
        /* send get address request (ping) */
        DS485_SendMessage(me, DS485_CreateMessage(me->pingDeviceAddress, me->deviceAddress, 0, DS485_GET_ADDRESS_REQUEST, 0));
        /* start timer */
        QTimeEvt_postIn(&me->registerStationTimeEvt, (QActive *)me, GET_ADDRESS_REQUEST_TIMEOUT_TICKS);
        return Q_HANDLED();
    }
    case Q_EXIT_SIG:
        /* stop timer */
        QTimeEvt_disarm(&me->registerStationTimeEvt);
        return Q_HANDLED();
    case REGISTER_STATION_TIMEOUT_SIG:
        /* get address request (ping) timed out */
        /* reset try counter */
        me->try = 0;
        return Q_TRAN(&DS485_send_set_address_request);
    case DS485_RX_GET_ADDRESS_RESPONSE_SIG:
    /* get address response received */
    if (DS485_GetSource((DS485MessageEvt const *)e) == me->pingDeviceAddress) {
        /* we got a ping response -> set address to used */
        SetBitSet8(me->usedAddresses, me->pingDeviceAddress);
        /* find next unused slave address */
        ++me->pingDeviceAddress;
        return Q_TRAN(DS485_ping_device_address);
    }
    break;
}
return Q_SUPER(&DS485_register_station);
}

/*..........................................................................*/
static QState DS485_send_set_address_request(DS485 *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG: {
        DS485MessageEvt* txMsgEvt;
        /* store history */
        me->tryRegisterStationHistory = (QStateHandler) &DS485_send_set_address_request;
        /* increment try counter */
        ++me->try;
        /* send set address request */
        txMsgEvt = DS485_CreateMessage(DS485_INTERMEDIATE_ADDRESS, me->deviceAddress, 0,
                DS485_SET_DEVICE_ADDRESS_REQUEST, 1);
        txMsgEvt->message[DS485_DATA_OFFSET] = me->pingDeviceAddress;
        DS485_SendMessage(me, txMsgEvt);
        /* start timer */
        QTimeEvt_postIn(&me->registerStationTimeEvt, (QActive *)me, SET_ADDRESS_RESPONSE_TIMEOUT_TICKS);
        return Q_HANDLED();
    }
    case Q_EXIT_SIG:
        QTimeEvt_disarm(&me->registerStationTimeEvt);
        return Q_HANDLED();
    case DS485_RX_SET_DEVICE_ADDRESS_RESPONSE_SIG:
        if (DS485_GetSource((DS485MessageEvt const *)e) == me->pingDeviceAddress) {
            /* disarm register station timer */
            QTimeEvt_disarm(&me->registerStationTimeEvt);
            /* reset try counter */
            me->try = 0;
            /* set successor address request */
            return Q_TRAN(&DS485_send_set_successor_address_request);
        }
        return Q_HANDLED();
    }
    return Q_SUPER(&DS485_register_station);
}

/*..........................................................................*/
static QState DS485_send_set_successor_address_request(DS485 *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG: {
        DS485MessageEvt* txMsgEvt;
        /* store history */
        me->tryRegisterStationHistory = (QStateHandler) &DS485_send_set_successor_address_request;
        /* increment try counter */
        ++me->try;
        /* send set successor address request */
        txMsgEvt = DS485_CreateMessage(me->pingDeviceAddress, me->deviceAddress, 0,
                DS485_SET_SUCCESSOR_ADDRESS_REQUEST, 1);
        /* successor address is me */
        txMsgEvt->message[DS485_DATA_OFFSET] = me->deviceAddress;
        DS485_SendMessage(me, txMsgEvt);
        /* start timer */
        QTimeEvt_postIn(&me->registerStationTimeEvt, (QActive *)me, SET_ADDRESS_RESPONSE_TIMEOUT_TICKS);
        return Q_HANDLED();
    }
    case Q_EXIT_SIG:
        QTimeEvt_disarm(&me->registerStationTimeEvt);
        return Q_HANDLED();
    case DS485_RX_SET_SUCCESSOR_ADDRESS_RESPONSE_SIG:
    {
        DS485MessageEvt const * rxMsgEvt = (DS485MessageEvt const *) e;
        /* check successor address response */
        if ((DS485_GetDestination(rxMsgEvt) == me->deviceAddress) && (DS485_GetSource(rxMsgEvt)
                == me->pingDeviceAddress)) {
            /* disarm register station timer */
            QTimeEvt_disarm(&me->registerStationTimeEvt);
            /* reset try counter */
            me->try = 0;
            return Q_TRAN(&DS485_send_set_successor_address_request_to_predecessor);
        }
        return Q_HANDLED();
    }
    }
    return Q_SUPER(&DS485_register_station);
}

/*..........................................................................*/
static QState DS485_send_set_successor_address_request_to_predecessor(DS485 *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG: {
        DS485MessageEvt* txMsgEvt;
        /* set history to exit */
        me->tryRegisterStationHistory = me->exitRegisterStationHistory;
        /* increment try counter */
        ++me->try;
        /* send set successor address request */
        txMsgEvt = DS485_CreateMessage(me->predecessorDeviceAddress, me->deviceAddress, 0,
                DS485_SET_SUCCESSOR_ADDRESS_REQUEST, 1);
        txMsgEvt->message[DS485_DATA_OFFSET] = me->pingDeviceAddress;
        DS485_SendMessage(me, txMsgEvt);
        /* start timer */
        QTimeEvt_postIn(&me->registerStationTimeEvt, (QActive *)me, SET_ADDRESS_RESPONSE_TIMEOUT_TICKS);
        return Q_HANDLED();
    }
    case Q_EXIT_SIG:
        QTimeEvt_disarm(&me->registerStationTimeEvt);
        return Q_HANDLED();
    case DS485_RX_SET_SUCCESSOR_ADDRESS_REQUEST_SIG:
    {
        DS485MessageEvt const * rxMsgEvt = (DS485MessageEvt const *) e;
        if (DS485_GetDestination(rxMsgEvt) == me->deviceAddress) {
            DS485MessageEvt* txMsgEvt;
            /* this device is the successor -> send set successor address response */
            me->successorDeviceAddress = rxMsgEvt->message[DS485_DATA_OFFSET];
            txMsgEvt = DS485_CreateMessage(DS485_GetSource(rxMsgEvt), me->deviceAddress, 0,
                    DS485_SET_SUCCESSOR_ADDRESS_RESPONSE, sizeof(me->dsmid));
            DS485_SetDSID(txMsgEvt, me->dsmid);
            DS485_SendMessage(me, txMsgEvt);
        }
        return Q_HANDLED();
    }
    case DS485_RX_SET_SUCCESSOR_ADDRESS_RESPONSE_SIG:
    {
        DS485MessageEvt const * rxMsgEvt = (DS485MessageEvt const *) e;
        if ((DS485_GetDestination(rxMsgEvt) == me->deviceAddress) && (DS485_GetSource(rxMsgEvt)
                == me->predecessorDeviceAddress)) {
            static const QEvent ds485DeviceRegistered =
            { MMI_DS485_DEVICE_REGISTERED_SIG, 0 };
            /* disarm register station timer */
            QTimeEvt_disarm(&me->registerStationTimeEvt);
            /* everything done -> set registered device address used */
            SetBitSet8(me->usedAddresses, me->pingDeviceAddress);
            /* set new predecessor address */
            me->predecessorDeviceAddress = me->pingDeviceAddress;
            /* send device registered to MMI */
            QActive_postFIFO(AO_MMI, &ds485DeviceRegistered);
            /* exit to history */
            return Q_TRAN(me->exitRegisterStationHistory);
        }
        return Q_HANDLED();
    }
    }
    return Q_SUPER(&DS485_register_station);
}

/*..........................................................................*/
static QState DS485_operational(DS485 *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG: {
        /* publish device address */
        DS485_PublishDeviceAddress(me->deviceAddress);
        /* notify ready to MMI */
        static const MMI_DS485StateEvt readyEvt = { { MMI_DS485_STATE_SIG, 0 }, MMI_DS485_READY };
        QActive_postFIFO(AO_MMI, (QEvent *)&readyEvt);
        ++me->statistics.enterOperational;
        /* initialize find station flag */
        me->findStation = me->master;
        return Q_HANDLED();
    }
    case Q_EXIT_SIG: {
        static const MMI_DS485StateEvt initializingEvt = { { MMI_DS485_STATE_SIG, 0 }, MMI_DS485_INITIALIZING };
        QActive_postFIFO(AO_MMI, (QEvent *)&initializingEvt);
        return Q_HANDLED();
    }
    case Q_INIT_SIG:
        if (me->master) {
            return Q_TRAN(&DS485_holding_token);
        }
        return Q_TRAN(&DS485_no_token);
    case FIND_STATION_TIMEOUT_SIG:
        me->findStation = TRUE;
        return Q_HANDLED();
    case SYSTEM_TERM_SIG:
        me->terminate = TRUE;
        return Q_HANDLED();
    case DS485_REQ_BUS_MEMBERS_SIG:
    {
        DS485_DsidResponseEvt    *evt    = Q_NEW(DS485_DsidResponseEvt,DS485_RES_BUS_MEMBERS_SIG);
        memcpy( evt->dsid, me->busDsid, sizeof(evt->dsid) );
        ccSendResponse( e, (QEvent*)evt );
        return Q_HANDLED();
    }
    }
    return Q_SUPER(&DS485_top);
}

/*..........................................................................*/
static QState DS485_holding_token(DS485 *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG:
        me->tokenMinHoldTimeElapsed = FALSE;
        me->tokenMaxHoldTimeElapsed = FALSE;
        /* start minimum hold timeout only if there are other devices on the bus */
        if ((me->deviceAddress == 0) && (me->successorDeviceAddress == 0)) {
            QTimeEvt_postIn(&me->tokenMaxTimeEvt, (QActive *)me, TOKEN_LONG_HOLD_TIMEOUT_TICKS);
        } else {
            QTimeEvt_postIn(&me->tokenMinTimeEvt, (QActive *)me, TOKEN_MIN_HOLD_TIMEOUT_TICKS);
            QTimeEvt_postIn(&me->tokenMaxTimeEvt, (QActive *)me, TOKEN_MAX_HOLD_TIMEOUT_TICKS);
        }
        me->messageSentAfterReceivingToken = FALSE;
        return Q_HANDLED();
    case Q_EXIT_SIG:
        QTimeEvt_disarm(&me->tokenMaxTimeEvt);
        return Q_HANDLED();
    case Q_INIT_SIG: {
        /* check for transmit data ready (internal message queue) */
        if (QActive_recall((QActive *) me, &me->deferredTxMessageQueue)) {
            return Q_TRAN(&DS485_transmit_data);
        }
        if (me->findStation) {
            return Q_TRAN(&DS485_find_station);
        }
        return Q_TRAN(&DS485_wait_for_token_min_hold_time);
    }
    case TOKEN_MIN_HOLD_TIMEOUT_SIG:
        me->tokenMinHoldTimeElapsed = TRUE;
        return Q_HANDLED();
    case TOKEN_MAX_HOLD_TIMEOUT_SIG:
        me->tokenMaxHoldTimeElapsed = TRUE;
        return Q_HANDLED();

    case DS485_RX_GET_ADDRESS_RESPONSE_SIG:
    {
        DS485ControlMessageEvt const * rxMsgEvt = (DS485ControlMessageEvt const *)e;
        if (DS485_GetDestination(rxMsgEvt) == me->deviceAddress) {
            printf("Register get-addr device %d\n", DS485_GetSource(rxMsgEvt));
            registerAddress(me, DS485_GetSource(rxMsgEvt), (dsid_t*)&rxMsgEvt->message[DS485_DATA_OFFSET]);
        }
        return Q_HANDLED();
    }
    case DS485_RX_RESPONSE_SIG:
    {
        DS485MessageEvt * rxMsgEvt = (DS485MessageEvt *) e;
        rxMsgEvt->sourceAddr = DS485_getDsidByBusId(me, DS485_GetSource(rxMsgEvt));
        if (DS485_GetMessageId(rxMsgEvt) == 0x91) {
            if (!isAddressValid(me,DS485_GetSource(rxMsgEvt))) {
                printf("Register api device %d\n", DS485_GetSource(rxMsgEvt));
                registerAddress(me, DS485_GetSource(rxMsgEvt), (dsid_t*)&rxMsgEvt->message[DS485_DATA_OFFSET+1]);
            }
        }
#if 0
        if (DS485_GetDestination(rxMsgEvt) == me->deviceAddress) {
            /* send acknowledge response and post message */
            DS485MessageEvt* txMsgEvt = DS485_CreateMessage(DS485_GetSource(rxMsgEvt), me->deviceAddress, 0,
                    DS485_ACKNOWLEDGE, 1);
            /* set message ID */
            txMsgEvt->message[DS485_DATA_OFFSET] = DS485_GetMessageId(rxMsgEvt);
            DS485_SendMessageEx(me, txMsgEvt);
            /* post message to DSM API active object */
            rxMsgEvt->destinationAddr = DS485_getDsidByBusId(me, DS485_GetDestination(rxMsgEvt));
            QActive_postFIFO(AO_DsmApi, e);
        }
#endif
        return Q_HANDLED();
    }

    }
    return Q_SUPER(&DS485_operational);
}

/*..........................................................................*/
static QState DS485_transmit_data(DS485 *me, QEvent const *e)
{
    switch (e->sig) {
    case DS485_TX_MSG_SIG: {
        bool isRequest;
        /* NOTE: cast away const! */
        DS485MessageEvt * txMsgEvt = (DS485MessageEvt *) e;
        /* set message sent after receiving token flag */
        me->messageSentAfterReceivingToken = TRUE;
        /* update source address */
        DS485_SetSource(txMsgEvt, me->deviceAddress);

        if( DS485_IsRequest(txMsgEvt) || DS485_IsResponse(txMsgEvt) || DS485_IsEvent(txMsgEvt)) {
            uint8_t dst;
            if( IsBroadcastId(txMsgEvt->destinationAddr) ){
                dst = 0x00;
                txMsgEvt->message[1] |= DS485_BROADCAST_MASK;
            } else {
                dst = DS485_getBusIdByDsid( me, &txMsgEvt->destinationAddr );
                if( dst == DS485_INVALID_ADDRESS ){
                    postTxResponse( -1 );
                    return  Q_HANDLED();
                }
            }
            DS485_SetDestination( txMsgEvt, dst );
        }

        /* send message */
        DS485_SendMessage(me, txMsgEvt);

        if( DS485_IsRequest(txMsgEvt) || DS485_IsResponse(txMsgEvt) || DS485_IsEvent(txMsgEvt)) {
            postTxResponse( 0 );
        }

        isRequest = DS485_IsRequest(txMsgEvt);
        if (isRequest) {
            ++me->statistics.txDataRequest;
        } else {
            ++me->statistics.txDataResponse;
        }
        /* is message broadcast or not a request (i.e. a response)? */
        if (DS485_IsBroadcast(txMsgEvt) || !isRequest) {
            /* recall next event */
            if (QActive_recall((QActive *) me, &me->deferredTxMessageQueue)) {
                return Q_HANDLED();
            }
            /* no more messages */
            return Q_TRAN(&DS485_wait_for_token_min_hold_time);
        }
        /* message is a unicast request -> start response timer */
        QTimeEvt_postIn(&me->transmitDataTimeEvt, (QActive *)me, DSMAPI_RESPONSE_TIMEOUT_TICKS);
        return Q_TRAN(&DS485_wait_for_acknowledge);
    }
    }
    return Q_SUPER(&DS485_holding_token);
}

/*..........................................................................*/
static QState DS485_wait_for_acknowledge(DS485 *me, QEvent const *e)
{
    switch (e->sig) {
    case DS485_TX_RESPONSE_SIG:
        /* fall through */
    case DS485_RX_BUSY_SIG:
        /* fall through */
    case DS485_RX_ACKNOWLEDGE_SIG:
    {
        DS485MessageEvt const * rxMsgEvt = (DS485MessageEvt const *) e;
        if (DS485_GetDestination(rxMsgEvt) == me->deviceAddress) {
            /* stop response timer */
            QTimeEvt_disarm(&me->transmitDataTimeEvt);
            /* post message to DSM API active object */
            QActive_postFIFO(AO_DsmApi, e);
            /* max hold time elapsed? */
            if (me->tokenMaxHoldTimeElapsed) {
                return Q_TRAN(&DS485_hand_over_token);
            }
            /* recall next event */
            if (QActive_recall((QActive *) me, &me->deferredTxMessageQueue)) {
                return Q_TRAN(&DS485_transmit_data);
            }
            /* min hold time elapsed? */
            if (me->tokenMinHoldTimeElapsed) {
                return Q_TRAN(&DS485_hand_over_token);
            }
            return Q_TRAN(&DS485_wait_for_token_min_hold_time);
        }
        return Q_HANDLED();
    }
    case DS485_RX_REQUEST_SIG:
        /* just ignore incoming requests */
        return Q_HANDLED();
    case TRANSMIT_DATA_TIMEOUT_SIG:
        /* timed out -> hand over token */
        return Q_TRAN(&DS485_hand_over_token);
    }
    return Q_SUPER(&DS485_holding_token);
}

/*..........................................................................*/
static QState DS485_find_station(DS485 *me, QEvent const *e)
{
    switch (e->sig)
    {
    case Q_ENTRY_SIG:
    {
        DS485MessageEvt* txMsgEvt;
        me->findStation = FALSE;
        /* send solicit successor long request */
        txMsgEvt = DS485_CreateMessage(0, me->deviceAddress, 0, DS485_SOLICIT_SUCCESSOR_REQUEST_LONG, sizeof(me->dsmid));
        DS485_SetDSID(txMsgEvt, me->dsmid);
        DS485_SendMessage(me, txMsgEvt);
        QTimeEvt_postIn(&me->responseTimeEvt, (QActive *)me, SOLICIT_SUCCESSOR_REQUEST_TIMEOUT_TICKS);
        /* restart find station timeout */
        QTimeEvt_postIn(&me->findStationTimeEvt, (QActive *)me, FIND_STATION_TIMEOUT_TICKS);
        return Q_HANDLED();
    }
    case DS485_RX_SOLICIT_SUCCESSOR_RESPONSE_SIG:
        me->responseTimeEvtDisarmed = QTimeEvt_disarm(&me->responseTimeEvt);
        me->superRegisterStationHistory = (QStateHandler) &DS485_holding_token;
        me->exitRegisterStationHistory = (QStateHandler) &DS485_hand_over_token;
        return Q_TRAN(&DS485_register_station);
    case RESPONSE_TIMEOUT_SIG:
        me->responseTimeEvtDisarmed = TRUE;
        return Q_TRAN(&DS485_hand_over_token);
    }
    return Q_SUPER(&DS485_holding_token);
}

/*..........................................................................*/
static QState DS485_wait_for_token_min_hold_time(DS485 *me, QEvent const *e)
{
    switch (e->sig) {
    case TOKEN_MIN_HOLD_TIMEOUT_SIG:
        /* fall through */
    case TOKEN_MAX_HOLD_TIMEOUT_SIG:
        return Q_TRAN(&DS485_hand_over_token);
    case DS485_RX_SOLICIT_SUCCESSOR_REQUEST_LONG_SIG:
        /* fall through */
    case DS485_RX_SOLICIT_SUCCESSOR_REQUEST_SHORT_SIG:
        /* fall through */
    case DS485_RX_GET_ADDRESS_REQUEST_SIG:
        /* fall through */
    case DS485_RX_SET_DEVICE_ADDRESS_REQUEST_SIG:
        /* fall through */
    case DS485_RX_SET_SUCCESSOR_ADDRESS_REQUEST_SIG:
        /* fall through */
#if 0
    case DS485_RX_EVENT_SIG:
        /* fall through */
    case DS485_RX_RESPONSE_SIG:
        /* fall through */
    case DS485_RX_REQUEST_SIG:
#endif
    {
        if (!me->messageSentAfterReceivingToken) {
            /* received a message which did not come from me -> restart */
            return Q_TRAN(&DS485_startup);
        }
    }
    break;

    case DS485_RX_TOKEN_SIG: {
        printf("wait_for_token_min_hold_time: received token again\n");
        if (DS485_GetDestination((DS485MessageEvt const *) e) != me->deviceAddress) {
            /* received a token, maybe due to repeated sends of predecessor, in any other case -> restart */
            return Q_TRAN(&DS485_startup);
        }
    }
    break;

    case DS485_TX_MSG_SIG:
        /* repost event */
        QActive_postLIFO((QActive *)me, e);
        return Q_TRAN(&DS485_transmit_data);
    }
    return Q_SUPER(&DS485_holding_token);
}

/*..........................................................................*/
static QState DS485_no_token(DS485 *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG:
        checkForMissingBusMembers(me);
        /* reset used addresses */
        memset(me->usedAddresses, 0, sizeof(me->usedAddresses));
        SetBitSet8(me->usedAddresses, me->deviceAddress);
        me->busDsid[me->deviceAddress]  = me->dsmid;
        /* start token missing timeout */
        QTimeEvt_postIn(&me->tokenMissingTimeEvt, (QActive *)me, (me->master ? TOKEN_MISSING_MASTER_TIMEOUT_TICKS : TOKEN_MISSING_SLAVE_TIMEOUT_TICKS));
        return Q_HANDLED();
    case Q_EXIT_SIG:
        QTimeEvt_disarm(&me->tokenMissingTimeEvt);
        return Q_HANDLED();
    case DS485_RX_TOKEN_SIG:
    {
        DS485MessageEvt const * rxMsgEvt = (DS485MessageEvt const *) e;
        /* update used addresses */
        if (!isAddressValid(me,DS485_GetSource(rxMsgEvt))) {
            DS485MessageEvt * txMsgEvt = DS485_CreateMessage(DS485_GetSource(rxMsgEvt), me->deviceAddress, 0, DS485_REQUEST, 1);
            /* Hack Alert: set message ID to dSM-API-Get-DSID */
            DS485_SetMsgID(txMsgEvt, 0x91);
            DS485_SendMessage(me, txMsgEvt);
        }
        SetBitSet8(me->usedAddresses, DS485_GetSource(rxMsgEvt));
        /* check destination */
        if (DS485_GetDestination(rxMsgEvt) == me->deviceAddress) {
            /* save predecessor address */
            me->predecessorDeviceAddress = DS485_GetSource(rxMsgEvt);
            /* holding token */
            return Q_TRAN(&DS485_holding_token);
        }
        return Q_HANDLED();
    }
    case DS485_RX_GET_ADDRESS_REQUEST_SIG:
    {
        DS485MessageEvt const * rxMsgEvt = (DS485MessageEvt const *) e;
        if (DS485_GetDestination(rxMsgEvt) == me->deviceAddress) {
            /* send get address response */
            DS485MessageEvt * txMsgEvt = DS485_CreateMessage(DS485_GetSource(rxMsgEvt), me->deviceAddress, 0,
                    DS485_GET_ADDRESS_RESPONSE, sizeof(me->dsmid));
            DS485_SetDSID(txMsgEvt, me->dsmid);
            DS485_SendMessage(me, txMsgEvt);
        }
        return Q_HANDLED();
    }
    case DS485_RX_GET_ADDRESS_RESPONSE_SIG:
    {
        DS485MessageEvt const * rxMsgEvt = (DS485MessageEvt const *) e;
        dsid_t id;
        if (DS485_GetPayloadLength(rxMsgEvt) == 4) {
            memcpy(&id, "\x35\x04\x17\x5f\xe0\x00\x00\x10\xff\xff\xff\xff", sizeof(dsid_t));
            memcpy(&id.id[8], &rxMsgEvt->message[DS485_DATA_OFFSET], 4);
        }
        else if (DS485_GetPayloadLength(rxMsgEvt) == sizeof(dsid_t)) {
            memcpy(&id, &rxMsgEvt->message[DS485_DATA_OFFSET], sizeof(dsid_t));
        }
        else {
            printf("Register from get-address-response: payload too short\n");
            break;
        }
        printf("Register get-address device %d\n", DS485_GetSource(rxMsgEvt));
        registerAddress(me, DS485_GetSource(rxMsgEvt), &id);
    }
    case DS485_RX_SET_SUCCESSOR_ADDRESS_REQUEST_SIG:
    {
        DS485MessageEvt const * rxMsgEvt = (DS485MessageEvt const *) e;
        if (DS485_GetDestination(rxMsgEvt) == me->deviceAddress) {
            /* save successor device address */
            me->successorDeviceAddress = rxMsgEvt->message[DS485_DATA_OFFSET];
            /* send set successor address response */
            DS485MessageEvt * txMsgEvt = DS485_CreateMessage(DS485_GetSource(rxMsgEvt), me->deviceAddress, 0,
                    DS485_SET_SUCCESSOR_ADDRESS_RESPONSE, sizeof(me->dsmid));
            DS485_SetDSID(txMsgEvt, me->dsmid);
            DS485_SendMessage(me, txMsgEvt);
        }
        return Q_HANDLED();
    }
    case DS485_RX_REQUEST_SIG:
    case DS485_RX_EVENT_SIG:
    {
        DS485MessageEvt * rxMsgEvt = (DS485MessageEvt *) e;
        rxMsgEvt->sourceAddr = DS485_getDsidByBusId(me, DS485_GetSource(rxMsgEvt));
        if (DS485_IsBroadcast(rxMsgEvt)) {
            /* broadcast request -> just post message */
            QActive_postFIFO(AO_DsmApi, e);
        }
        else if (DS485_GetDestination(rxMsgEvt) == me->deviceAddress) {
            /* unicast request -> send acknowledge */
            DS485MessageEvt * txMsgEvt = DS485_CreateMessage(DS485_GetSource(rxMsgEvt), me->deviceAddress, 0,
                    DS485_ACKNOWLEDGE, 1);
            /* set message ID */
            txMsgEvt->message[DS485_DATA_OFFSET] = DS485_GetMessageId(rxMsgEvt);
            DS485_SendMessageEx(me, txMsgEvt);
            /* post message to DSM API active object */
            QActive_postFIFO(AO_DsmApi, e);
        }
        return Q_HANDLED();
    }
    case DS485_RX_ACKNOWLEDGE_SIG:
        /* fall through */
    case DS485_RX_BUSY_SIG:
    {
        DS485MessageEvt * rxMsgEvt = (DS485MessageEvt *) e;
        rxMsgEvt->sourceAddr = DS485_getDsidByBusId(me, DS485_GetSource(rxMsgEvt));
        if (DS485_GetDestination(rxMsgEvt) == me->deviceAddress) {
            /* post message to DSM API active object */
            QActive_postFIFO(AO_DsmApi, e);
        }
        return Q_HANDLED();
    }
    case TOKEN_MISSING_TIMEOUT_SIG:
        if (me->master) {
            /* this device is master -> generate and hand over new token */
            return Q_TRAN(&DS485_hand_over_token);
        }
        /* this device is slave -> start from scratch */
        return Q_TRAN(&DS485_startup);
    }
    return Q_SUPER(&DS485_operational);
}

/*..........................................................................*/
static QState DS485_hand_over_token(DS485 *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG:
        me->tokenHandoverCount = 0;
        return Q_HANDLED();
    case Q_INIT_SIG:
        return Q_TRAN(&DS485_send_token);
    }
    return Q_SUPER(&DS485_operational);
}

/*..........................................................................*/
static QState DS485_send_token(DS485 *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG:
        ++me->tokenHandoverCount;
        /* send token to successor device */
        DS485_SendToken(me, me->successorDeviceAddress, me->deviceAddress);
        QTimeEvt_postIn(&me->handOverTokenTimeEvt, (QActive *)me, TOKEN_HAND_OVER_TIMEOUT_TICKS);
        return Q_HANDLED();
    case Q_EXIT_SIG:
        QTimeEvt_disarm(&me->handOverTokenTimeEvt);
        return Q_HANDLED();
    case HAND_OVER_TOKEN_TIMEOUT_SIG:
        if (me->tokenHandoverCount == 3) {
            /* reset hand over count */
            me->tokenHandoverCount = 0;
            /* successor device lost */
            QActive_postFIFO(AO_MMI, &l_ds485DeviceLostEvt);
            /* check if unavailable successor is master */
            if (me->successorDeviceAddress == DS485_MASTER_ADDRESS) {
                /* master did not respond -> I am master now */
                DS485_ChangeToMaster(me);
            }
            /* get next successor station */
            if (DS485_GetNextSuccessorAddress(me)) {
                /* find successor device */
                return Q_TRAN(&DS485_find_successor_device);
            }
            /* just re-send token */
        }
        /* re-send token */
        return Q_TRAN(&DS485_send_token);
    case DS485_RX_TOKEN_SIG:
        /* fall through */
    case DS485_RX_SOLICIT_SUCCESSOR_REQUEST_LONG_SIG:
        /* fall through */
    case DS485_RX_SOLICIT_SUCCESSOR_REQUEST_SHORT_SIG:
        /* fall through */
    case DS485_RX_GET_ADDRESS_REQUEST_SIG:
        /* fall through */
    case DS485_RX_SET_DEVICE_ADDRESS_REQUEST_SIG:
        /* fall through */
    case DS485_RX_SET_SUCCESSOR_ADDRESS_REQUEST_SIG:
        /* fall through */
    case DS485_RX_RESPONSE_SIG:
        /* fall through */
    case DS485_RX_EVENT_SIG:
        /* fall through */
    case DS485_RX_REQUEST_SIG: {
        /* check source address */
        DS485MessageEvt const * rxMsgEvt = (DS485MessageEvt const *) e;
        if (DS485_GetSource(rxMsgEvt) == me->successorDeviceAddress) {
            /* successor is active */
            if (!me->terminate) {
                /* repost message */
                QActive_postLIFO((QActive *)me, e);
                /* change to no token state */
                return Q_TRAN(&DS485_no_token);
            }
            return Q_TRAN(&DS485_stopped);
        }
        return Q_HANDLED();
    }
    }
    return Q_SUPER(&DS485_hand_over_token);
}

/*..........................................................................*/
static QState DS485_find_successor_device(DS485 *me, QEvent const *e)
{
    switch (e->sig)
    {
    case Q_ENTRY_SIG:
        /* send get address request (ping) */
        DS485_SendMessage(me, DS485_CreateMessage(me->successorDeviceAddress, me->deviceAddress, 0,
                DS485_GET_ADDRESS_REQUEST, 0));
        /* start timer */
        QTimeEvt_postIn(&me->handOverTokenTimeEvt, (QActive *)me, GET_ADDRESS_REQUEST_TIMEOUT_TICKS);
        return Q_HANDLED();
    case Q_EXIT_SIG:
        /* stop timer */
        QTimeEvt_disarm(&me->handOverTokenTimeEvt);
        return Q_HANDLED();
    case DS485_RX_GET_ADDRESS_RESPONSE_SIG:
        /* get address response received */
        if (DS485_GetSource((DS485MessageEvt const *)e) == me->successorDeviceAddress) {
            /* we got a ping response -> send token to this new successor */
            return Q_TRAN(&DS485_send_token);
        }
        return Q_HANDLED();
    case HAND_OVER_TOKEN_TIMEOUT_SIG:
        /* check if unavailable successor is master */
        if (me->successorDeviceAddress == DS485_MASTER_ADDRESS) {
            /* master did not respond -> I am master now */
            DS485_ChangeToMaster(me);
        }
        if (DS485_GetNextSuccessorAddress(me)) {
            return Q_TRAN(&DS485_find_successor_device);
        }
        /* I am the successor -> no other devices on the bus */
        return Q_TRAN(&DS485_send_token);
    }
    return Q_SUPER(&DS485_hand_over_token);
}

/*..........................................................................*/
static QState DS485_stopped(DS485 *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG: {
        /* send DS485 stopped event */
        static const QEvent ds485StoppedEvt = { DS485_STOPPED_SIG, 0 };
        QActive_postFIFO(AO_MMI, &ds485StoppedEvt);
        return Q_HANDLED();
    }
    }
    return Q_SUPER(&QHsm_top);
}

/*..........................................................................*/
static void DS485_SendToken(DS485* me, uint8_t dest, uint8_t src)
{
    DS485MessageEvt* msgEvt = Q_NEW(DS485MessageEvt, DS485_RX_TOKEN_SIG);
    uint8_t* msg = msgEvt->message;
    /* build header */
    msg[0] = DS485_START_BYTE;
    msg[1] = (dest << 2);
    msg[2] = (src << 2);
    /* set message size */
    msgEvt->size = 3;
    /* post message to queue */
    QEQueue_postLIFO(&l_ds485TxMessageQueue, (QEvent const *) msgEvt);
    ds485_system_tx();
}

/*..........................................................................*/
static void DS485_SendMessage(DS485 *me, DS485MessageEvt *msgEvt)
{
    uint8_t * message = msgEvt->message;
    /* calculate crc16 ccitt */
    uint16_t crc = crc16_ccitt_calc_data(0, message, msgEvt->size - 2);
    /* add crc16 (least significant byte first) */
    message[msgEvt->size - 2] = (uint8_t) (crc & 0xFF);
    message[msgEvt->size - 1] = (uint8_t) (crc >> 8);
    /* post message to queue */
    QEQueue_postFIFO(&l_ds485TxMessageQueue, (QEvent const *) msgEvt);
    ds485_system_tx();
}

/*..........................................................................*/
static void DS485_SendMessageEx(DS485 *me, DS485MessageEvt *msgEvt)
{
    uint8_t * message = msgEvt->message;
    /* calculate crc16 ccitt */
    uint16_t crc = crc16_ccitt_calc_data(0, message, msgEvt->size - 2);
    /* add crc16 (least significant byte first) */
    message[msgEvt->size - 2] = (uint8_t) (crc & 0xFF);
    message[msgEvt->size - 1] = (uint8_t) (crc >> 8);
    /* post message to queue */
    QEQueue_postLIFO(&l_ds485TxMessageQueue, (QEvent const *) msgEvt);
    ds485_system_tx();
}

/*..........................................................................*/
static void DS485_PublishDeviceAddress(uint8_t deviceAddress)
{
    DS485AddressEvt * addrEvt = Q_NEW(DS485AddressEvt, DS485_ADDRESS_SIG);
    addrEvt->deviceAddress = deviceAddress;
    QF_publish((QEvent *) addrEvt);
}

/*..........................................................................*/
static void DS485_ChangeToMaster(DS485 *me)
{
    me->master = TRUE;
    me->findStation = TRUE;
    ClearBitSet8(me->usedAddresses, me->deviceAddress);
    me->busDsid[me->deviceAddress]  = NO_DEVICE;
    unregisterAddress( me, DS485_MASTER_ADDRESS );        // lost master
    me->deviceAddress = DS485_MASTER_ADDRESS;
    SetBitSet8(me->usedAddresses, me->deviceAddress);
    me->busDsid[me->deviceAddress]  = me->dsmid;
    /* publish device address */
    DS485_PublishDeviceAddress(me->deviceAddress);
}

/*..........................................................................*/
static bool DS485_GetNextSuccessorAddress(DS485 *me)
{
    do {
        /* find next successor station */
        if (++me->successorDeviceAddress > DS485_MAX_SLAVE_ADDRESS) {
            me->successorDeviceAddress = DS485_MIN_ADDRESS;
        }
    } while ( IsBitSet8Cleared(me->usedAddresses,me->successorDeviceAddress));
    if (me->successorDeviceAddress != me->deviceAddress) {
        return TRUE;
    }
    return FALSE;
}

/*..........................................................................*/
static void DS485_ResetStatistics(DS485 *me)
{
    memset(&me->statistics, 0, sizeof(me->statistics));
}

/*..........................................................................*/
static void DS485_RequestStatistics(DS485 *me, DS485ReqStatisticsEvt const * reqEvt)
{
    DS485ResStatisticsEvt * resEvt = Q_NEW(DS485ResStatisticsEvt, DS485_RES_STATISTICS_SIG);
    memcpy(&resEvt->statistics, &me->statistics, sizeof(resEvt->statistics));
    QActive_postFIFO(reqEvt->requester, (QEvent *)resEvt);
}

static uint8_t DS485_getBusIdByDsid( DS485 *me, dsid_t *dsid )
{
    int i;
    for( i = 0; i < sizeof(me->busDsid)/sizeof(me->busDsid[0]); i++ ){
        if( cmpAddress(me,i,dsid) == 0 ){
            return  i;
        }
    }
    return  DS485_INVALID_ADDRESS;
}

static dsid_t DS485_getDsidByBusId( DS485 *me, uint8_t busId )
{
    return me->busDsid[busId];
}
